/*
 EasyPen Input Server Add-on. EasyPen.ccp Ver 1.0 
 Copyright (C) 2001  Robin Taylor.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; version 2 of the License.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details
*/

#include "EasyPen.h"

//Instance new device
BInputServerDevice* instantiate_input_device()
{
 return(new EasyPenDevice());
}

//EasyPen device constructor
EasyPenDevice::EasyPenDevice( void ):BInputServerDevice()
{
 input_device_ref dev = { "Genius Easypen Tablet", B_POINTING_DEVICE, NULL };
 input_device_ref *dev_list[2] = { &dev, NULL };
 RegisterDevices( dev_list );
}

//EasyPen destructor
EasyPenDevice::~EasyPenDevice(void)
{
 //Delete BSerialPort object
 delete serial;
}
//Initialization check

status_t EasyPenDevice::InitCheck()
{
 char DeviceName[B_OS_NAME_LENGTH],Byte[]="1";
 int SerialPorts;
 //Create a new BSerialPort object
 serial = new BSerialPort();  
 //Count serial ports
 SerialPorts = serial->CountDevices()-1;
 //Check all the ports for an EasyPen,(Summa Device)
 for(;SerialPorts>=0; SerialPorts-- )
	{//get the name of port number SerialPorts
	 serial->GetDeviceName( SerialPorts, DeviceName );
	 //configure the port to 9600,O,8,1 	        
     serial->SetDataRate( B_9600_BPS );
	 serial->SetDataBits( B_DATA_BITS_8 );
	 serial->SetStopBits( B_STOP_BITS_1 );
     serial->SetParityMode( B_ODD_PARITY );
	 serial->SetFlowControl(B_HARDWARE_CONTROL+B_SOFTWARE_CONTROL);
	 serial->SetBlocking( true );
     serial->SetTimeout( 800000 ); 
     //open the port
	 if(serial->Open(DeviceName)>0)
       { 
		 //wait 0.25s let everything settle       
         snooze( 250000 );
		 //Set serial comms to 8bits and no parity
		 serial->Write("z9", 2 );
		 //send NULL for RESET
		 serial->Write((char)NULL, 1 );
         //Send Remote request D mode P trigger
         serial->Write("DP", 2 );
         //wait 0.25s
         snooze( 250000 );
         //Clear input buffer
  		 serial->ClearInput();
         //Run DIAGNOSTIC (t) test and RETURN (w) results trigger (P)
         serial->Write("twP", 3 );
  	     //Read a Byte 0.3s timeout on the port...
  	     serial->Read((void *)&(Byte[0]), 1 );
         //Did the device send a good diagnostic byte?
         //Mask the proximaty BIT good test should alway then give 87h
         if((Byte[0]&0xF7)!=0x87) continue;//If not FAIL try next port
	     //Return B_OK if the input server is B_OK
  	     return(BInputServerDevice::InitCheck());
  	     } 
      //(Loop scope)
      }	      
 //No EasyPen or Summa class tablet found!
 //Don't load the driver.	      
 return(B_ERROR);
}

//Return driver thread to start
int32 easypen_thread( void* device )
{ //returns a pointer to the 'thread' function
  return(((EasyPenDevice*)device)->EasyThread());
}

//spawn a new thread
status_t EasyPenDevice::Start( const char *device, void * )
{
  Quit = false;
  threadid = spawn_thread(easypen_thread, device,B_DISPLAY_PRIORITY, this );
  resume_thread( threadid );
  return(B_NO_ERROR);
}

//stop thread
status_t EasyPenDevice::Stop( const char *, void * )
{
  Quit = true;
  status_t err = B_OK;
  wait_for_thread( threadid, &err );
  threadid = B_ERROR;
  return(B_NO_ERROR);
}

//Main 'Driver' Thread
int32 EasyPenDevice::EasyThread(void)
{
 BMessage* event;
 bigtime_t ClickSpeed, TimeNow, LastClicked=0;
 char Byte[6];
 float X_Axis_f, Y_Axis_f;
 unsigned int Read,X,Y,Total,Button,OldButton=0,ClickCount=0;
 
  // ':' Set Easypen Binary message format Summa is 'zb'
  // '@' Set Stream mode/
  // 'F' Set Absolute coordinates
  // 'R' Set Data rate Maximum+2
  // 'b' Set Y Origin upper left
  serial->Write( ":@FRb", 5 );
  //get double click speed
  get_click_speed( &ClickSpeed );

  while(!Quit )
       {
		//(Re)Set Read byte to reinitiated the get 1st byte loop
		Read=0;  
		while(!Read) Read = serial->Read((void *)&Byte[0], 1);
		//If bit7 0x80 Phasing bit NOT set or bit6 Proximity bit set restart main loop
		if((Byte[0]&0x40 || !(Byte[0]&0x80))) continue;
		//Set bytes to read for second loop
		Total=4;
		while(Total)
		     {
		       Read = serial->Read((void *)&Byte[5-Total], Total);
		       Total-=Read;
		       }
		//If any of the four bytes have bit7 set the data is invalid,restart main loop     
		if(Byte[1]&0x80 || Byte[2]&0x80 || Byte[3]&0x80 || Byte[4]&0x80) continue;

		//Get Button info by masking relevant bits
		Button=Byte[0]&7;
		//Make X and Y data Words
		X = ((unsigned int)(Byte[2] << 7) | (unsigned int)Byte[1]);
		Y = ((unsigned int)(Byte[4] << 7) | (unsigned int)Byte[3]);	
		// Convert X and Y into floating point positions
        X_Axis_f = (float)((double)X / (double)2000.0);
 	    Y_Axis_f = (float)((double)Y / (double)1496.0);

		//Get time now
		TimeNow = system_time();
		//Set the BMessage type dependent on any changes in button state
        if(Button==OldButton) event = new BMessage( B_MOUSE_MOVED );
           else if(Button>OldButton)
                  {
                   event = new BMessage( B_MOUSE_DOWN );
                   //if not a 'double click' count=1 else add the click
                   if(LastClicked + ClickSpeed <= TimeNow) ClickCount=1;
                     else ClickCount++;
                   //remeber when this click occured  
                   LastClicked = TimeNow; 
                   event->AddInt32( "clicks", ClickCount);
                   }
                    else event = new BMessage( B_MOUSE_UP );
  
 		//Save the current but state for comparison
        OldButton=Button;
        //make the rest of the message
        event->AddInt64( "when", TimeNow);
		event->AddFloat( "x", X_Axis_f);
		event->AddFloat( "y", Y_Axis_f);
		event->AddInt32("buttons", Button);
		event->AddInt32("modifiers", modifiers());
		//and for other programs...
		event->AddFloat( "be:tablet_x", X_Axis_f );
		event->AddFloat( "be:tablet_y", Y_Axis_f );
  	    //Post the message
		EnqueueMessage( event );
		
     	}
 //'Cause it expect a return value... it isn't used :)     	
 return(B_OK);
}